/****************************************************************************
 *
 * amph1: Beispiel zur Mehrfachvererbung
 *
 *  - Beispielprogramm zum Buch "Objektorientiertes Programmieren in C++",
 *    von Nicolai Josuttis, erschienen bei Addison-Wesley, ISBN 3-89319-637-4
 *
 *  - Abschnitt 4.5.1: Beispiel zur Mehrfachvererbung
 *  
 ****************************************************************************/

/****************************************************************************
 *
 * Copyright 1994 by Nicolai Josuttis and Addison-Wesley.
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software for personal
 * and educational use is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice
 * and this permission notice appear in supporting documentation, and that
 * the names of Addison-Wesley or the authors are not be used in advertising
 * or publicity pertaining to distribution of the software without specific,
 * written prior permission.  Addison-Wesley and the authors make no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * ADDISON-WESLEY AND THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL ADDISON-WESLEY OR THE AUTHORS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 ****************************************************************************/

// da die meisten Compiler "bool" (noch) nicht koennen (siehe Seite 43):
typedef int bool;
#define true 1
#define false 0



/****************************************************************************
 *
 *	Auto.h
 *
 ****************************************************************************/

#ifndef _Auto_h
#define _Auto_h

// Header-Datei fuer I/O einbinden
#include <iostream.h>

/* Klasse Auto
 *  - zur Vererbung geeignet
 */
class Auto {
  protected:
    int km;        // gefahrene Kilometer

  public:
    // Default- und int-Konstruktor
    Auto (int d = 0) {
        km = d;         // gefahrene Kilometer initialisieren
    }

    // bestimmte Anzahl von Kilometern fahren
    virtual void fahre (int d) {
        km += d;        // Kilometer aufaddieren
    }

    // Anzahl gefahrener Kilometer ausgeben
    virtual void printGefahren () {
        cout << "Das Auto ist " << km << " km gefahren" << endl;
    }

    // virtueller Destruktor (ohne Anweisungen)
    virtual ~Auto () {
    }
};

#endif    /* _Auto_h */


/****************************************************************************
 *
 *	Boot.h
 *
 ****************************************************************************/

#ifndef _Boot_h
#define _Boot_h

// Header-Datei fuer I/O einbinden
#include <iostream.h>

/* Klasse Boot
 *  - zur Vererbung geeignet
 */
class Boot {
  protected:
    int sm;        // gefahrene Seemeilen

  public:
    // Default- und int-Konstruktor
    Boot (int d = 0) {
        sm = d;         // gefahrene Seemeilen initialisieren
    }

    // bestimmte Anzahl von Seemeilen fahren
    virtual void fahre (int d) {
        sm += d;        // Seemeilen aufaddieren
    }

    // Anzahl gefahrener Seemeilen ausgeben
    virtual void printGefahren () {
        cout << "Das Boot ist " << sm << " sm gefahren" << endl;
    }

    // virtueller Destruktor (ohne Anweisungen)
    virtual ~Boot () {
    }
};

#endif    /* _Boot_h */


/****************************************************************************
 *
 *	Amph.h
 *
 ****************************************************************************/

#ifndef _Amph_h
#define _Amph_h

// Header-Dateien der Basisklassen einbinden
//#include "Auto.h"
//#include "Boot.h"

/* Klasse Amph
 *  - abgeleitet von Auto und Boot
 *  - zur Weiter-Vererbung geeignet
 */
class Amph : public Auto, public Boot {
  public:
    /* Default-, int und int/int-Konstruktor
     *  - mit erstem Parameter wird Auto-Konstruktor aufgerufen
     *  - mit zweitem Parameter wird Boot-Konstruktor aufgerufen
     */
    Amph (int k = 0, int s = 0) : Auto(k), Boot(s) {
        // damit ist nichts mehr zu tun
    }

    // Anzahl gefahrener Kilometer und Seemeilen ausgeben
    virtual void printGefahren () {
        cout << "Das Amphibienfahrzeug ist " << km << " km und "
             << sm << " sm gefahren" << endl;
    }

    // virtueller Destruktor (ohne Anweisungen)
    virtual ~Amph () {
    }
};

#endif    /* _Amph_h */


/****************************************************************************
 *
 *	atest1.cc / atest1.cpp
 *
 ****************************************************************************/

// Header-Datei fuer die Klasse Amph
//#include "Amph.h"

main ()
{
    /* Amphibienfahrzeug anlegen und mit
     * 7 Kilometer und 42 Seemeilen initialisieren
     */
    Amph a (7, 42);

    // gefahrene Kilometer und Seemeilen ausgeben
    a.printGefahren();
}

